package com.mmbnetworks.configuration;

import com.mmbnetworks.serialdriver.SerialDriver;
import com.mmbnetworks.serial.rha.ARHAFrame;
import com.mmbnetworks.serial.rha.utility.RHAHostStartupReady;
import com.mmbnetworks.serial.rha.utility.RHASerialACKConfigWrite;
import com.mmbnetworks.serial.rha.utility.RHAStartupSyncComplete;
import com.mmbnetworks.serial.rha.utility.RHAStartupSyncRequest;
import com.mmbnetworks.serial.rha.utility.RHAStatusResponse;
import com.mmbnetworks.serial.rha.zigbeesupportconfig.RHAAddEndpoint;
import com.mmbnetworks.serial.rha.zigbeesupportconfig.RHAClearEndpointConfig;
import com.mmbnetworks.serial.rha.zigbeesupportconfig.RHADeviceTypeWrite;
import com.mmbnetworks.serial.rha.zigbeesupportconfig.RHARegisterCommandsPassThru;
import com.mmbnetworks.serial.types.ClusterID;
import com.mmbnetworks.serial.types.ConfigurationStateEnum;
import com.mmbnetworks.serial.types.RunningStateEnum;
import com.mmbnetworks.serial.types.SerialAckConfigEnum;
import com.mmbnetworks.serial.types.StatusEnum;
import com.mmbnetworks.serial.types.UInt16;
import com.mmbnetworks.serial.types.UInt8;
import com.mmbnetworks.serial.types.ZigBeeArray;

import java.util.ArrayList;
import java.util.List;

public class StartupSyncStateMachine {
    
    final static private UInt8 LOCAL_ENDPOINT = new UInt8((short)0x01); 

    private enum State {
        WAITING_STARTUP_SYNC_REQUEST,
        SERIAL_ACK_ENABLE,
        SERIAL_ACK_ENABLE_ACK,
        HOST_STARTUP_READY,
        HOST_STARTUP_READY_ACK,
        CONFIGURE_DEVICE_TYPE,
        CONFIGURE_DEVICE_TYPE_ACK,
        CLEAR_ENDPOINTS,
        CLEAR_ENDPOINTS_ACK,
        ADD_ENDPOINTS_AND_CLUSTERS,
        ADD_ENDPOINTS_AND_CLUSTERS_ACK,
        ADD_ATTRIBUTES,
        ADD_ATTRIBUTES_ACK,
        UPDATE_ATTRIBUTES,
        UPDATE_ATTRIBUTES_ACK,
        OTHER_CONFIG,
        OTHER_CONFIG_ACK,
        STARTUP_SYNC_COMPLETE,
        STARTUP_SYNC_COMPLETE_ACK,

        NUM_STATES
    }
    private State mCurrentState;
    private int mFrameSequenceNumber;
    private final SerialDriver mSerialDriver;

    public StartupSyncStateMachine(final SerialDriver serialDriver) {
        mCurrentState = State.WAITING_STARTUP_SYNC_REQUEST;
        mFrameSequenceNumber = 0xFF;
        mSerialDriver = serialDriver;
    }

    public void handleFrame(final ARHAFrame rhaFrame) {

        switch (mCurrentState) {
            case WAITING_STARTUP_SYNC_REQUEST:
                if (rhaFrame instanceof RHAStartupSyncRequest) {
                    transitionTo(State.SERIAL_ACK_ENABLE);
                }
                break;
            case SERIAL_ACK_ENABLE_ACK:
                if (isAck(rhaFrame, mFrameSequenceNumber)) {
                    transitionTo(State.HOST_STARTUP_READY);
                }
                break;
            case HOST_STARTUP_READY_ACK:
                // the acknowledgment of a HOST_STARTUP_READY is a start up sync request frame
                if (rhaFrame instanceof RHAStartupSyncRequest) {
                    final RHAStartupSyncRequest frame = (RHAStartupSyncRequest)rhaFrame;
                    boolean requiresFullConfig = false;
                    boolean requiresEndpointConfig = false;
                    // now configure
                    if (frame.getRunningState().getValue() == RunningStateEnum.ConcreteRunningStateEnum.STARTING_UP.getCode()) {
                        if (frame.getConfigurationState().getValue() == ConfigurationStateEnum.ConcreteConfigurationStateEnum.FACTORY_DEFAULT.getCode()) {
                            requiresFullConfig = true;
                        } else if (frame.getConfigurationState().getValue() == ConfigurationStateEnum.ConcreteConfigurationStateEnum.NEEDS_ENDPOINT_CONFIGURATION.getCode()) {
                            requiresEndpointConfig = true;
                        } else if (frame.getConfigurationState().getValue() == ConfigurationStateEnum.ConcreteConfigurationStateEnum.FULLY_CONFIGURED.getCode()) {
                            // do nothing, will be handled below
                        } else {
                            // just do full reconfiguration if unknown state
                            requiresFullConfig = true;
                        }
                        
                        // transition to proper state based on current application status
                        if (requiresFullConfig) {
                            transitionTo(State.CONFIGURE_DEVICE_TYPE);
                        } else if (requiresEndpointConfig) {
                            transitionTo(State.CLEAR_ENDPOINTS);
                        } else {
                            transitionTo(State.STARTUP_SYNC_COMPLETE);
                        }
                        
                    } else {
                        // if running state is already running
                        transitionTo(State.STARTUP_SYNC_COMPLETE);
                    }
                }
                break;
            case CONFIGURE_DEVICE_TYPE_ACK:
                if (isAck(rhaFrame, mFrameSequenceNumber)) {
                    transitionTo(State.CLEAR_ENDPOINTS);
                }
                break;
            case CLEAR_ENDPOINTS_ACK:
                if (isAck(rhaFrame, mFrameSequenceNumber)) {
                    transitionTo(State.ADD_ENDPOINTS_AND_CLUSTERS);
                }
                break;
            case ADD_ENDPOINTS_AND_CLUSTERS_ACK:
                if (isAck(rhaFrame, mFrameSequenceNumber)) {
                    // Note: we skip directly to other config
                    // See comments in transitionTo for more information about how to implement
                    // the ADD_ATTRIBUTES state if it is required
                    //transitionTo(ADD_ATTRIBUTES);
                    transitionTo(State.OTHER_CONFIG);
                }
                break;
            case ADD_ATTRIBUTES_ACK:
                if (isAck(rhaFrame, mFrameSequenceNumber)) {
                    transitionTo(State.UPDATE_ATTRIBUTES);
                }
                break;
            case UPDATE_ATTRIBUTES_ACK:
                if (isAck(rhaFrame, mFrameSequenceNumber)) {
                    transitionTo(State.OTHER_CONFIG);
                }
                break;
            case OTHER_CONFIG_ACK:
                if (isAck(rhaFrame, mFrameSequenceNumber)) {
                    transitionTo(State.STARTUP_SYNC_COMPLETE);
                }
                break;
            case STARTUP_SYNC_COMPLETE_ACK:
                if (isAck(rhaFrame, mFrameSequenceNumber)) {
                    transitionTo(State.WAITING_STARTUP_SYNC_REQUEST);
                }
                break;
            default:
                System.out.println("StartupSyncStateMachine::receiveMessage - Unhandled State " + mCurrentState);
                break;
        }
    }

    private void transitionTo(final State nextState) {
        mCurrentState = nextState;

        // moving to the following states requires some action to be taken
        switch (nextState) {
            case WAITING_STARTUP_SYNC_REQUEST:
                mFrameSequenceNumber = 0;
                break;
            case SERIAL_ACK_ENABLE:
            {
                final RHASerialACKConfigWrite frame = new RHASerialACKConfigWrite();
                frame.setSerialConfig(new SerialAckConfigEnum((byte)0x01));
                mFrameSequenceNumber = sendMessage(frame);
                mCurrentState = State.SERIAL_ACK_ENABLE_ACK;
            }
                break;
            case HOST_STARTUP_READY:
            {
                final RHAHostStartupReady frame = new RHAHostStartupReady();
                mFrameSequenceNumber = sendMessage(frame);
                mCurrentState = State.HOST_STARTUP_READY_ACK;
            }
                break;
            case CONFIGURE_DEVICE_TYPE:
            {
                final RHADeviceTypeWrite frame = new RHADeviceTypeWrite();
                frame.setDeviceFunctionType(new UInt8((short)0x00)); // full-functional device
                frame.setSleepy(new UInt8((short)0x00)); // non-sleepy device
                mFrameSequenceNumber = sendMessage(frame);
                mCurrentState = State.CONFIGURE_DEVICE_TYPE_ACK;
            }
                break;
            case CLEAR_ENDPOINTS:
            {
                final RHAClearEndpointConfig frame = new RHAClearEndpointConfig();
                mFrameSequenceNumber = sendMessage(frame);
                mCurrentState = State.CLEAR_ENDPOINTS_ACK;
            }
                break;
            case ADD_ENDPOINTS_AND_CLUSTERS:
            {
                mFrameSequenceNumber = endpointConfig();
                mCurrentState = State.ADD_ENDPOINTS_AND_CLUSTERS_ACK;
            }
                break;
            case ADD_ATTRIBUTES:
                // not implemented; could be used if you want to add attributes that the RapidConnect
                // firmware doesn't automatically add when a cluster is added
                break;
            case UPDATE_ATTRIBUTES:
                // not implemented; could be used if you want to update attributes to your own
                // default values - perhaps such as the Manufacturer Name attribute on the Basic cluster
                break;
            case OTHER_CONFIG:
            {
                // register passthru on the IAS Zone client cluster to pass through Zone Alarm commands
                final RHARegisterCommandsPassThru frame = new RHARegisterCommandsPassThru();
                frame.setEndpointID(LOCAL_ENDPOINT); 
                frame.setClusterID(new ClusterID(0x0500)); // ias zone
                frame.setClusterServerClient(new UInt8((short)0x00)); // client cluster

                // command id list to pass through
                final List<UInt8> clist = new ArrayList<UInt8>();
                clist.add(new UInt8((short)0x00)); // ias zone status change
                final ZigBeeArray<UInt8> commandList = new ZigBeeArray<UInt8>(UInt8.class, 1, clist);
                frame.setCommandIDsList(commandList);

                mFrameSequenceNumber = sendMessage(frame);
                mCurrentState = State.OTHER_CONFIG_ACK;
            }
                break;
            case STARTUP_SYNC_COMPLETE:
            {
                final RHAStartupSyncComplete frame = new RHAStartupSyncComplete();
                mFrameSequenceNumber = sendMessage(frame);
                mCurrentState = State.STARTUP_SYNC_COMPLETE_ACK;
            }
                break;
            default:
                System.out.println("StartUpSyncStateMachine::transitionTo - Unhandled State " + nextState);
                break;
        }
    }
    
    private boolean isAck(final ARHAFrame rhaFrame, final int frameSeq) {
        boolean retVal = false;

        if (rhaFrame instanceof RHAStatusResponse) {
            final RHAStatusResponse frame = (RHAStatusResponse) rhaFrame;
            retVal = (frame.getFrameSequence() == frameSeq)
                    && (frame.getStatus().getValue() == StatusEnum.ConcreteStatusEnum.SUCCESS.getCode());
        }

        return retVal;
    }
    
    private int sendMessage(final ARHAFrame frame) {
        return mSerialDriver.sendMessage(frame);
    }

    @SuppressWarnings({ "unchecked", "rawtypes" })
    private int endpointConfig() {
        final RHAAddEndpoint f1 = new RHAAddEndpoint();
        f1.setEndpointID(LOCAL_ENDPOINT);
        f1.setProfileID(new UInt16(0x104));
        f1.setDeviceID(new UInt16(0x0007));
        f1.setDeviceVersion(new UInt8((short) 0x01));

        // server clusters
        final List<ClusterID> serverClusterIDList = new ArrayList();
        final int[] sclist = {
            0x0000, // basic
            0x0003  // identify
        };
        for (int i = 0; i < sclist.length; i++) {
            final ClusterID s = new ClusterID(sclist[i]);
            serverClusterIDList.add(s);
        }
        final ZigBeeArray<ClusterID> serverClusterArray = new ZigBeeArray(ClusterID.class, 1, serverClusterIDList);
        f1.setServerClusterList(serverClusterArray);

        // client clusters
        final List<ClusterID> clientClusterIDList = new ArrayList();
        final int[] cclist = {
            0x0000, // basic
            0x0003, // identify
            0x0004, // groups
            0x0005, // scenes
            0x0006, // on/off
            0x0008, // level control
            0x0101, // door lock
            0x0201, // thermostat
            0x0202, // fan control
            0x0500  // ias zone
        };
        for (int i = 0; i < cclist.length; i++) {
            final ClusterID c = new ClusterID(cclist[i]);
            clientClusterIDList.add(c);
        }
        final ZigBeeArray<ClusterID> clientClusterArray = new ZigBeeArray(ClusterID.class, 1, clientClusterIDList);
        f1.setClientClusterList(clientClusterArray);

        return sendMessage(f1);
    }

}
