package com.mmbnetworks.host;

import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;

import com.mmbnetworks.clustercommands.ClusterCommands;
import com.mmbnetworks.configuration.StartupSyncStateMachine;
import com.mmbnetworks.devicetable.DeviceTable;
import com.mmbnetworks.serial.rha.ARHAFrame;
import com.mmbnetworks.serial.rha.networkcomissioning.RHAFormNetwork;
import com.mmbnetworks.serial.rha.networkcomissioning.RHAJoinNetwork;
import com.mmbnetworks.serial.rha.networkcomissioning.RHALeaveNetwork;
import com.mmbnetworks.serial.rha.networkcomissioning.RHANetworkStatusRequest;
import com.mmbnetworks.serial.rha.networkcomissioning.RHAPermitJoin;
import com.mmbnetworks.serial.rha.utility.RHAModuleInfoRequest;
import com.mmbnetworks.serial.rha.utility.RHARestoreDefaults;
import com.mmbnetworks.serial.rha.zdomessages.RHAZDODeviceAnnounceReceived;
import com.mmbnetworks.serial.types.Bitmap32;
import com.mmbnetworks.serial.types.Bitmap8;
import com.mmbnetworks.serial.types.ExtendedPanId;
import com.mmbnetworks.serial.types.NodeId;
import com.mmbnetworks.serial.types.PanId;
import com.mmbnetworks.serial.types.UInt8;
import com.mmbnetworks.serialdriver.SerialDriver;
import com.mmbnetworks.upgrade.ISerialUpgradeNotification;
import com.mmbnetworks.upgrade.SerialUpgradeStateMachine;

// This example file shows how to setup a serial driver, and begin sending and receiving frames.
// The Host sends frames to the Module via the SerialDriver.sendMessage function.
// The Host receives messages from the Module in the EventHandler.run() method, which is a single-threaded event loop.
// There is an example of a simple state machine to configure the module.
// There is an example of a simple state machine to serially upgrade the module.

public class Host implements ISerialUpgradeNotification {

    private final AtomicBoolean mEventLoopRunning = new AtomicBoolean(false);
    private final BlockingQueue<Object> mEventQueue = new LinkedBlockingQueue<Object>();
    private final DeviceTable mDeviceTable;
    private final SerialDriver mSerialDriver;
    private final SerialUpgradeStateMachine mSerialUpgradeStateMachine;
    private final Thread mEventHandlerThread;

    public Host(final String portName) throws HostInitializationException {
        mSerialDriver = new SerialDriver(portName);
        boolean connected = mSerialDriver.connect();
        if (!connected) {
            throw new HostInitializationException("Problem Connecting to port" + portName);
        }
        mSerialDriver.registerEventQueue(mEventQueue);

        mSerialUpgradeStateMachine = new SerialUpgradeStateMachine(mSerialDriver);
        mSerialUpgradeStateMachine.setCallback(this);
        mDeviceTable = new DeviceTable();

        // create and start the event handler thread
        mEventHandlerThread = new Thread(new EventHandler());
        mEventHandlerThread.start();
    }

    public void disconnect() {
        try {
            mSerialUpgradeStateMachine.setCallback(null);
            mEventLoopRunning.set(false);
            mEventHandlerThread.interrupt();
            mEventHandlerThread.join(5000); // ms
            mSerialDriver.close();
        } catch (Throwable th) {
            // ignore
        }
    }

    public void moduleInfoRequest() {
        final RHAModuleInfoRequest frame = new RHAModuleInfoRequest();
        mSerialDriver.sendMessage(frame);
    }
    
    public void networkStatusRequest() {
        final RHANetworkStatusRequest frame = new RHANetworkStatusRequest();
        mSerialDriver.sendMessage(frame);
    }
    
    public void joinNetwork() {
        final RHAJoinNetwork frame = new RHAJoinNetwork();
        mSerialDriver.sendMessage(frame);
    }
    
    public void leaveNetwork() {
        final RHALeaveNetwork frame = new RHALeaveNetwork();
        mSerialDriver.sendMessage(frame);
    }
    
    public void formNetwork() {
        final RHAFormNetwork req = new RHAFormNetwork();

        // set all channels
        Bitmap32 channelMask = new Bitmap32();
        for (int i = 11; i < 27; i++) {
            channelMask.setBit(i);
        }
        req.setChannelMask(channelMask);

        // auto-everything
        req.setAutoOptions(new Bitmap8(new byte[] {0x03}));
        req.setShortPANID(new PanId());
        req.setExtendedPANID(new ExtendedPanId());

        // send message
        mSerialDriver.sendMessage(req);
    }
    
    public void permitJoin(short duration) {
        final RHAPermitJoin req = new RHAPermitJoin();
        req.setJoinDuration(new UInt8(duration));
        mSerialDriver.sendMessage(req);
    }
    
    public void restoreFactoryDefaults() {
        final RHARestoreDefaults req = new RHARestoreDefaults();
        mSerialDriver.sendMessage(req);
    }
    
    public void startSerialUpgrade(final String filePath) {
        mSerialUpgradeStateMachine.startUpgrade(filePath);
    }
        
    public void sendOnCommand(NodeId nodeId, short endpointId) {
        mSerialDriver.sendMessage(ClusterCommands.getOnCommandFrame(nodeId,endpointId));
    }
    
    public void sendOffCommandFrame(NodeId nodeId, short endpointId) {
        mSerialDriver.sendMessage(ClusterCommands.getOffCommandFrame(nodeId,endpointId));
    }
    
    public void sendToggleCommandFrame(NodeId nodeId, short endpointId) {
        mSerialDriver.sendMessage(ClusterCommands.getToggleCommandFrame(nodeId,endpointId));
    }
    
    @Override
    public void notifyUpgradeSuccess() {
        System.out.println("Serial Upgrade Succeeded");
    }

    @Override
    public void notifyUpgradeFailed(int errorCode) {
        System.out.println("Serial Upgrade Failed: " + errorCode);
    }

    private final class EventHandler implements Runnable {
         
        private final StartupSyncStateMachine mStartupSyncStateMachine; 

        public EventHandler() {
            mStartupSyncStateMachine = new StartupSyncStateMachine(mSerialDriver);
        }
        
        @Override
        public void run() {

            mEventLoopRunning.set(true);
            final long TIMEOUT_SEC = 20;

            // single-threaded event loop that will handle messages off the serial handler
            while (mEventLoopRunning.get()) {
                try {
                    // will block here until there's something in the queue, or until a specified timeout
                    final Object o = mEventQueue.poll(TIMEOUT_SEC, TimeUnit.SECONDS);

                    // process timeouts
                    if (o == null) {
                        mSerialUpgradeStateMachine.handleTimeout();
                        continue;
                    }

                    // otherwise, handle frames
                    if (o instanceof ARHAFrame) {
                        // make sure that these run mutually exclusively
                        if (mSerialUpgradeStateMachine.isRunning()) {
                            mSerialUpgradeStateMachine.handleFrame((ARHAFrame) o);
                        } else {
                            mStartupSyncStateMachine.handleFrame((ARHAFrame) o);
                        }
                    }

                    if (o instanceof RHAZDODeviceAnnounceReceived) {
                        mDeviceTable.handleDeviceAnnounce((RHAZDODeviceAnnounceReceived) o);
                    }

                } catch (InterruptedException th) {
                    // ignore
                }
            }

            System.out.println("Quitting EventHandler...");
        }
    }

}
