package com.mmbnetworks.serialdriver;

import com.mmbnetworks.serial.IFrame;
import com.mmbnetworks.serial.ISerialDelegate;
import com.mmbnetworks.serial.MMBParser;
import com.mmbnetworks.serial.rha.ARHAFrame;
import com.mmbnetworks.serial.rha.otabootload.RHAOTAImageBlockResponse;
import com.mmbnetworks.serial.rha.otabootload.RHAOTAImageNotification;
import com.mmbnetworks.serial.rha.otabootload.RHAOTAQueryNextImageResponse;
import com.mmbnetworks.serial.rha.otabootload.RHAOTAUpgradeEndResponse;
import com.mmbnetworks.serial.rha.utility.RHAHostStartupReady;
import com.mmbnetworks.serial.rha.utility.RHAReset;
import com.mmbnetworks.serial.rha.utility.RHARestoreDefaults;
import com.mmbnetworks.serial.rha.utility.RHASerialACKConfigResponse;
import com.mmbnetworks.serial.rha.utility.RHASerialACKConfigWrite;
import com.mmbnetworks.serial.rha.utility.RHAStatusResponse;
import com.mmbnetworks.serial.rha.zigbeesupportconfig.RHAAttributeReportPassthroughControl;
import com.mmbnetworks.serial.types.SerialAckConfigEnum;
import com.mmbnetworks.serial.types.StatusEnum;

import java.util.Iterator;
import java.util.concurrent.BlockingDeque;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.LinkedBlockingDeque;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.logging.Level;
import java.util.logging.Logger;

import jssc.SerialPort;
import jssc.SerialPortEvent;
import jssc.SerialPortEventListener;
import jssc.SerialPortException;

@SuppressWarnings({ "unchecked", "rawtypes" })
public class SerialDriver implements ISerialDelegate {

    //-------------------------------------------------------------------------
    //    Public Fields
    //-------------------------------------------------------------------------

    //-------------------------------------------------------------------------
    //    Private Fields
    //-------------------------------------------------------------------------
    // variables needed for sending/receiving messages and acks
    private final AtomicBoolean mSerialAckEnabled;
    private final BlockingDeque<WriteObject> mOutgoingMessages;
    private final BlockingQueue<ARHAFrame> mAckSignal;
    private BlockingQueue mIncomingMessages = null;
    private final ConcurrentMap<Byte, IFrame> mSerialAckMap;
    private static byte mNextSequenceNumber = 0;
    private static final Object mNextSequenceNumberLock = new Object();

    // parsing and framing
    private final MMBParser mParser;

    // serial ports and threads
    private SerialPort mSerialPort = null;
    private SerialPortWriter mSerialWriter = null;
    private final String mConnectionName;
    private Thread mWriteThread = null;

    // various constants
    private static final int SERIAL_ACK_TIMEOUT_MS = 750;
    private static final int SERIAL_ACK_RETRY_COUNT = 2;

    //-------------------------------------------------------------------------
    //    Public Methods
    //-------------------------------------------------------------------------
    public SerialDriver(final String connectionName) {
        mSerialAckEnabled = new AtomicBoolean(false);
        mOutgoingMessages = new LinkedBlockingDeque();
        mAckSignal = new LinkedBlockingQueue<>();
        mSerialAckMap = new ConcurrentHashMap<>();
        mParser = new MMBParser(this);
        mConnectionName = connectionName;
    }

    public synchronized boolean connect() {
        mSerialPort = new SerialPort(mConnectionName);

        // check if port is in use
        if (mSerialPort.isOpened()) {
            mSerialPort = null;
            log("Error: Port " + mConnectionName + " is currently in use");
            return false;
        } else {
            try {
                mSerialPort.openPort();

                log("Open Serial Port: " + mConnectionName);

                // Serial Port Settings
                mSerialPort.setParams(115200, SerialPort.DATABITS_8, SerialPort.STOPBITS_1, SerialPort.PARITY_NONE, false, false);
                mSerialPort.setEventsMask(SerialPort.MASK_RXCHAR);
                mSerialPort.addEventListener(new SerialPortReader(), SerialPort.MASK_RXCHAR);

                // Set up the thread to handle serial output
                mSerialWriter = new SerialPortWriter(mSerialPort);
                mWriteThread = new Thread(mSerialWriter, "TXTHREAD" + mConnectionName);
                mWriteThread.start();

                log("Done connecting.");

                setSerialAck(true);

                return true;
            } catch (SerialPortException e) {
                log(String.format("Failed to connect on %s:  %s", mConnectionName, e.toString()));
                return false;
            }
        }
    }

    public synchronized void close() {
        if (mSerialPort != null) {
            log(String.format("Close Serial Port: %s", mSerialPort.getPortName()));

            try {
                if (mSerialWriter != null) {
                    mSerialWriter.stop();
                }
                if (mWriteThread != null) {
                    mWriteThread.interrupt();
                    mWriteThread.join();
                }
            } catch (InterruptedException e) {
                log(String.format("Error while closing %s:  %s", mConnectionName, e.toString()));
            } finally {
                mWriteThread = null;
            }
            try {
                if (mSerialPort.isOpened()) {
                    mSerialPort.closePort();
                }
                mSerialPort = null;
            } catch (SerialPortException e) {
                log(String.format("Error while closing %s:  %s", mConnectionName, e.toString()));
            }

            try {
                mOutgoingMessages.clear();
                mAckSignal.clear();
                mSerialAckMap.clear();
            } catch (UnsupportedOperationException ex) {
                // ignore
            }

        }

        log("Done close.");
    }

    public synchronized void registerEventQueue(final BlockingQueue eventQueue) {
        mIncomingMessages = eventQueue;
    }

    @Override
    public void receiveMessage(IFrame c) {
        // This will get called from our mParser object, which is called from the SerialPortReader thread
        log(String.format("RX: %s :[ %s]", c.getClass().getSimpleName(), SerialUtil.bytesToHex(c.getBytes())));

        // set the serial ack state if required
        if (c instanceof RHASerialACKConfigResponse) {
            final RHASerialACKConfigResponse ackState = (RHASerialACKConfigResponse) c;
            final boolean ackEnabled = (ackState.getSerialConfig().getValue() == SerialAckConfigEnum.ConcreteSerialAckConfigEnum.SERIAL_ACK_ENABLED.getCode());
            mSerialAckEnabled.set(ackEnabled);
        }

        // Handle the Serial ACK feature if enabled
        if (mSerialAckEnabled.get()) {
            final ARHAFrame receivedFrame = (ARHAFrame) c;

            // If there's stuff in the mSerialAckMap, then we signal
            // the write thread so that it can proceed to send the ACK (if required)
            if (!mSerialAckMap.isEmpty()) {
                mAckSignal.offer(receivedFrame);
            } else {
                // otherwise, que the ack to the front of the write thread
                if (frameRequiresAckFromHost(receivedFrame)) {
                    final RHAStatusResponse ack = new RHAStatusResponse();
                    ack.setFrameSequence(receivedFrame.getFrameSequence());
                    ack.setStatus(new StatusEnum(StatusEnum.ConcreteStatusEnum.SUCCESS.getCode()));

                    final WriteObject o = new WriteObject(ack);
                    mOutgoingMessages.offerFirst(o);
                }
            }
        }

        // Finally, pass the received frame to registered listeners
        if (mIncomingMessages != null) {
            mIncomingMessages.offer(c);
        }
    }

    public byte sendMessage(IFrame c) {
        // This will queue up a message for the SerialPortWriter thread to process,
        // and return the frame sequence number that was used for the message.
        byte returnVal = (byte)0xFF;

        if (mSerialPort != null) {
            if (c instanceof ARHAFrame) {
                final ARHAFrame rhaFrame = (ARHAFrame)c;
                returnVal = getNextSequenceNumber();
                rhaFrame.setFrameSequence(returnVal);
            }
            final WriteObject o = new WriteObject(c);
            mOutgoingMessages.offer(o);
        } else {
            log("Error sending message; no serial port connected.");
        }

        return returnVal;
    }

    public void setSerialAck(final boolean enabled) {
        if (mSerialPort != null) {
            final RHASerialACKConfigWrite ackConfig = new RHASerialACKConfigWrite();
            if (enabled) {
                ackConfig.setSerialConfig(new SerialAckConfigEnum((byte)1));
            } else {
                ackConfig.setSerialConfig(new SerialAckConfigEnum((byte)0));
            }

            sendMessage(ackConfig);
        } else {
            log("Error: Serial Port is not connected.");
        }
    }

    //-------------------------------------------------------------------------
    //    Protected Methods
    //-------------------------------------------------------------------------
    // This is just for debugging
    protected void debugPrintAllQueues() {
        final Iterator<WriteObject> writeMessages = mOutgoingMessages.iterator();
        while (writeMessages.hasNext()) {
            final WriteObject o = writeMessages.next();
            System.out.println("Print Queue - Write Queue: " + o.frame.getClass().getSimpleName());
        }

        final Iterator<ARHAFrame> ackSignals = mAckSignal.iterator();
        while (ackSignals.hasNext()) {
            final ARHAFrame f = ackSignals.next();
            System.out.println("Print Queue - Ack Signals: " + f.getClass().getSimpleName());
        }

        final Iterator<IFrame> serialAckMap = mSerialAckMap.values().iterator();
        while (serialAckMap.hasNext()) {
            final IFrame f = serialAckMap.next();
            System.out.println("Print Queue - Serial Ack Map: " + f.getClass().getSimpleName());
        }
    }

    //-------------------------------------------------------------------------
    //    Private Methods
    //-------------------------------------------------------------------------
    private void log(final String msg) {
        // TODO: replace with your own logging system
        System.out.println(msg);
    } 

    private boolean frameRequiresAckFromModule(final IFrame c) {
        // These are the exceptions; if a frame is not listed here,
        // then it requires an ACK (or associated response) from the Module
        return !(  c instanceof RHAStatusResponse
                || c instanceof RHAHostStartupReady
                || c instanceof RHAReset
                || c instanceof RHARestoreDefaults
                || c instanceof RHAOTAImageNotification
                || c instanceof RHAOTAQueryNextImageResponse
                || c instanceof RHAOTAImageBlockResponse
                || c instanceof RHAOTAUpgradeEndResponse
                || c instanceof RHAAttributeReportPassthroughControl);
    }

    private boolean frameRequiresAckFromHost(final IFrame c) {
        // These are the exceptions; if a frame is not listed here,
        // then it requires an ACK from the Host
        return !(c instanceof RHAStatusResponse);
    }

    //-------------------------------------------------------------------------
    //    Private Classes
    //-------------------------------------------------------------------------
    // This class receives events from the Rx thread
    private class SerialPortReader implements SerialPortEventListener {

        @Override
        public void serialEvent(SerialPortEvent event) {
            // we're not interested in anything that isn't a received char
            if (!event.isRXCHAR()) {
                return;
            }

            final int numBytes = event.getEventValue();
            if (numBytes > 0) {
                try {
                    final byte buffer[] = mSerialPort.readBytes(numBytes);
                    mParser.receiveBytes(buffer);
                } catch (SerialPortException ex) {
                    log(String.format("Error reading bytes on %s", mConnectionName, ex.toString()));
                }
            }
        }
    }

    // This class runs on its own thread and is used to write to the serial port
    private class SerialPortWriter implements Runnable {

        private final SerialPort mOutput;
        private final AtomicBoolean mKeepAlive = new AtomicBoolean(true);

        public SerialPortWriter(final SerialPort output) {
            this.mOutput = output;
        }

        @Override
        public void run() {

            while (mKeepAlive.get()) {
                try {
                    final WriteObject o = mOutgoingMessages.take(); // will block here until there's something in the queue
                    if (o != null) {
                        final IFrame frame = o.frame;

                        // We need to set mSerialAckEnabled here. If set from other threads, then you have a
                        // race condition where this variable can be knocked out of sync with the Module.
                        // We also want to wait until we receive an ACK for this before sending other frames.
                        final boolean isSerialAckConfig = (frame instanceof RHASerialACKConfigWrite);
                        if (isSerialAckConfig) {
                            final RHASerialACKConfigWrite ackConfig = (RHASerialACKConfigWrite)frame;
                            final boolean ackEnabled = ackConfig.getSerialConfig().getValue() == SerialAckConfigEnum.ConcreteSerialAckConfigEnum.SERIAL_ACK_ENABLED.getCode();
                            mSerialAckEnabled.set(ackEnabled);
                            if (!ackEnabled) {
                                mSerialAckMap.clear();
                            }
                        }

                        // we call "getBytes" so that the frame sequence number and timestamp get set
                        final byte[] rawBytes = frame.getBytes();

                        // Que the frame to wait for an ACK before we send it. If you do it elsewhere, a response
                        // to this frame may come in before the mSerialAckMap had had a chance to record this frame.
                        if (mSerialAckEnabled.get() && frameRequiresAckFromModule(frame)) {
                            mSerialAckMap.put(((ARHAFrame)frame).getFrameSequence(), frame);
                        }

                        // send the frame
                        this.mOutput.writeBytes(rawBytes);
                        log(String.format("TX: %s :[ %s]", frame.getClass().getSimpleName(), SerialUtil.bytesToHex(rawBytes)));

                        // now wait for an ACK or timeout before proceeding to the next frame in the queue
                        if (mSerialAckEnabled.get() && frameRequiresAckFromModule(frame)) {

                            while (!mSerialAckMap.isEmpty()) {

                                // will block here for a certain timeout, or until there's something on the queue
                                final ARHAFrame receivedFrame = mAckSignal.poll(SERIAL_ACK_TIMEOUT_MS, TimeUnit.MILLISECONDS);

                                if (receivedFrame == null) {
                                    // it was a timeout, so retry 
                                    if (o.retryCount < SERIAL_ACK_RETRY_COUNT) {
                                        o.retryCount++;
                                        mOutgoingMessages.offerFirst(o);
                                    } else {
                                        // We never got a response or an ACK to this after several retries,
                                        // so just remove it.
                                        // TODO: propagate error up to application?
                                        if (isSerialAckConfig) {
                                            // we tried to enable serial ack and it failed,
                                            // so reset state variable to what it was
                                            mSerialAckEnabled.set(!mSerialAckEnabled.get());
                                        }
                                        mSerialAckMap.remove(((ARHAFrame)o.frame).getFrameSequence());
                                    }

                                    break; // while (!mSerialAckMap.isEmpty())
                                }

                                // Otherwise, we received a frame:
                                // It's either a frame from the Module that needs to be ACK'd,
                                // or, it's an ACK (or response) from the Module to something we sent out previously

                                // First, ACK the frame from the module if it's a frame that should be ACK'd
                                if ( frameRequiresAckFromHost(receivedFrame) ) {
                                    final RHAStatusResponse ack = new RHAStatusResponse();
                                    ack.setFrameSequence(receivedFrame.getFrameSequence());
                                    ack.setStatus(new StatusEnum(StatusEnum.ConcreteStatusEnum.SUCCESS.getCode()));

                                    this.mOutput.writeBytes(ack.getBytes());
                                    log(String.format("TX: %s :[ %s]", ack.getClass().getSimpleName(), SerialUtil.bytesToHex(ack.getBytes())));
                                }

                                // If we were writing a serial ack config, check to see if it worked. If not, reset state variable.
                                if (isSerialAckConfig && receivedFrame instanceof RHAStatusResponse) {
                                    final RHAStatusResponse sr = (RHAStatusResponse)receivedFrame;
                                    final boolean success = ( sr.getStatus().getValue() == StatusEnum.ConcreteStatusEnum.SUCCESS.getCode() );
                                    if (!success) {
                                        mSerialAckEnabled.set(!mSerialAckEnabled.get());
                                    }
                                }

                                // Finally, check to see if this frame is a response to a frame that we sent out previously.
                                // If so, remove it from our serial ack queue so we can proceed writing other frames
                                mSerialAckMap.remove(receivedFrame.getFrameSequence());
                            }

                        }
                    }
                } catch (InterruptedException ex) {
                    // ignore
                } catch (SerialPortException | NullPointerException ex) {
                    Logger.getLogger(SerialDriver.class.getName()).log(Level.SEVERE, null, ex);
                }
            }
        }

        public void stop() {
            mKeepAlive.set(false);
        }
    }

    private static byte getNextSequenceNumber() {
        synchronized(mNextSequenceNumberLock) {
            byte returnValue = mNextSequenceNumber;
            mNextSequenceNumber++;
            if (mNextSequenceNumber > 127) {
                mNextSequenceNumber = 0;
            }
            return returnValue;
        }
    }

    private class WriteObject {
        public int retryCount = 0;
        public final IFrame frame;

        public WriteObject(final IFrame f) {
            this.frame = f;
        }
    }

}
