package com.mmbnetworks.upgrade;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.Calendar;
import java.util.TimeZone;

import com.mmbnetworks.serial.rha.ARHAFrame;
import com.mmbnetworks.serial.rha.bootload.RHAImageBlockRequest;
import com.mmbnetworks.serial.rha.bootload.RHAImageBlockResponse;
import com.mmbnetworks.serial.rha.bootload.RHAQueryNextImageResponse;
import com.mmbnetworks.serial.rha.bootload.RHAUpgradeEndRequest;
import com.mmbnetworks.serial.rha.bootload.RHAUpgradeEndResponse;
import com.mmbnetworks.serial.rha.networkcomissioning.RHANetworkStatusResponse;
import com.mmbnetworks.serial.rha.utility.RHAError;
import com.mmbnetworks.serial.rha.utility.RHAModuleInfoRequest;
import com.mmbnetworks.serial.rha.utility.RHAModuleInfoResponse;
import com.mmbnetworks.serial.rha.utility.RHASerialACKConfigWrite;
import com.mmbnetworks.serial.rha.utility.RHAStartupSyncComplete;
import com.mmbnetworks.serial.rha.utility.RHAStartupSyncRequest;
import com.mmbnetworks.serial.rha.utility.RHAStatusResponse;
import com.mmbnetworks.serial.types.IEEEAddress;
import com.mmbnetworks.serial.types.NodeId;
import com.mmbnetworks.serial.types.OctetString;
import com.mmbnetworks.serial.types.SerialAckConfigEnum;
import com.mmbnetworks.serial.types.UInt16;
import com.mmbnetworks.serial.types.UInt32;
import com.mmbnetworks.serial.types.UInt8;
import com.mmbnetworks.serial.types.UTCTime;
import com.mmbnetworks.serial.types.ZCLStatusEnum;
import com.mmbnetworks.serialdriver.SerialDriver;

public class SerialUpgradeStateMachine {

    private enum State {
        STATE_INIT,
        STATE_STARTUP_SYNC_COMPLETE,
        STATE_STARTUP_SYNC_RESPONSE,
        STATE_MODULE_INFO_REQUEST,
        STATE_MODULE_INFO_RESPONSE,
        STATE_START_TRANSFER,
        STATE_IN_TRANSFER,
        STATE_PERFORM_UPGRADE,
        STATE_COMPLETE
    };
    
    public SerialUpgradeStateMachine(final SerialDriver serialDriver) {
        mCallback = null;
        mSerialDriver = serialDriver;
        
        resetInternalState();
    }

    public void setCallback(ISerialUpgradeNotification n) {
        mCallback = n;
    }
    
    public boolean startUpgrade(final String filePath) {
        // check to ensure we're in the right state
        if (mState != State.STATE_INIT) {
            handleUpgradeFailure(ISerialUpgradeNotification.SERIAL_UPGRADE_FAILED_WRONG_START_STATE, false);
            return false;
        }

        try {
            // open the file
            File file = new File(filePath);
            mFileSize = file.length();
            mFirmwareFile = new RandomAccessFile(file, "r");
            System.out.println("Using file: " + filePath);
            System.out.println("File size: " + mFileSize);

        } catch (Throwable th) {
            System.out.println("Error opening file at: " +  filePath);
            handleUpgradeFailure(ISerialUpgradeNotification.SERIAL_UPGRADE_FAILED_ERROR_OPENING_FILE, true);
            return false;
        }

        // enter next state to start the upgrade
        transitionTo(State.STATE_STARTUP_SYNC_COMPLETE);

        return true;
    }

    public void handleFrame(final ARHAFrame c) {
        if (handleErrorFrame(c)) {
            return;
        }

        switch (mState) {
        case STATE_STARTUP_SYNC_RESPONSE:
            if ( c instanceof RHANetworkStatusResponse
              || c instanceof RHAStatusResponse ) {
                transitionTo(State.STATE_MODULE_INFO_REQUEST);
            }
            break;
        case STATE_MODULE_INFO_RESPONSE:
            if (c instanceof RHAModuleInfoResponse) {
                final RHAModuleInfoResponse frame = (RHAModuleInfoResponse)c;
                mEui64 = frame.getEui64();

                transitionTo(State.STATE_START_TRANSFER);
            }
            break;
        case STATE_IN_TRANSFER:
            if (c instanceof RHAImageBlockRequest) {
                // get the data size we should use
                final RHAImageBlockRequest frame = (RHAImageBlockRequest)c;
                final short maxDataSize = frame.getMaximumDataSize().getValue();
                final long offset = frame.getFileOffset().getValue();
                int bufferSize = maxDataSize;
                // correct the size if what is being requested exceeds the file
                if ((offset + maxDataSize) > mFileSize) {
                    bufferSize = (int)(mFileSize - offset);
                }
                // get the file data from the offset and copy into local buffer
                byte[] buffer = new byte[bufferSize];
                try {
                    mFirmwareFile.seek(offset);
                    mFirmwareFile.read(buffer, 0, bufferSize);
                } catch (Throwable th) {
                    abortUpgrade();
                    handleUpgradeFailure(ISerialUpgradeNotification.SERIAL_UPGRADE_FAILED_LOCAL_FILE_ERROR, false);
                    return;
                }

                // form a response
                final RHAImageBlockResponse resp = new RHAImageBlockResponse();
                resp.setNodeId(frame.getNodeId());
                resp.setEui64(frame.getEui64());
                resp.setEndpoint(frame.getEndpoint());
                resp.setStatus(new ZCLStatusEnum(ZCLStatusEnum.ConcreteZCLStatusEnum.SUCCESS.getCode()));
                resp.setManufacturerCode(frame.getManufacturerCode());
                resp.setImageType(frame.getImageType());
                resp.setFileVersion(frame.getFileVersion());
                resp.setFileOffset(frame.getFileOffset());
                OctetString frameData = new OctetString();
                frameData.setValue(buffer);
                resp.setData(frameData);

                mSerialDriver.sendMessage(resp);

            } else if (c instanceof RHAUpgradeEndRequest) {
                final RHAUpgradeEndRequest frame = (RHAUpgradeEndRequest)c;
                if (frame.getStatus().getValue() == UPGRADE_END_REQUEST_SUCCESS) {
                    final RHAUpgradeEndResponse resp = new RHAUpgradeEndResponse();
                    resp.setNodeId(frame.getNodeId());
                    resp.setEui64(frame.getEui64());
                    resp.setEndpoint(frame.getEndpoint());
                    resp.setManufacturerCode(frame.getManufacturerCode());
                    resp.setImageType(frame.getImageType());
                    resp.setFileVersion(frame.getFileVersion());
                    Calendar tmpCalendar = Calendar.getInstance(TimeZone.getTimeZone("GMT"));
                    tmpCalendar.set(2000, 0, 1, 0, 0, 0);
                    tmpCalendar.set(Calendar.MILLISECOND, 0);
                    resp.setCurrentTime(new UTCTime(tmpCalendar.getTime()));
                    resp.setUpgradeTime(new UTCTime(tmpCalendar.getTime()));

                    mSerialDriver.sendMessage(resp);

                    transitionTo(State.STATE_PERFORM_UPGRADE);
                }
            }
            break;
        case STATE_PERFORM_UPGRADE:
            if (c instanceof RHAStartupSyncRequest) {
                transitionTo(State.STATE_COMPLETE);
            }
            break;
        case STATE_COMPLETE:
            if (c instanceof RHAStatusResponse) {
                final RHAStatusResponse frame = (RHAStatusResponse)c;
                if ((frame.getFrameSequence().intValue() == mAckSequenceNumber)
                 && (frame.getStatus().getValue() == (byte)0x00) ) {
                    transitionTo(State.STATE_INIT);
                    handleUpgradeSuccess();
                }
            }
            break;
        default:
            break;
        }
    }
    
    public void handleTimeout() {
        if (mState != State.STATE_INIT) {
            handleUpgradeFailure(ISerialUpgradeNotification.SERIAL_UPGRADE_FAILED_ERROR_TIMEOUT, true);
        }
    }

    public boolean isRunning() {
        return (mState != State.STATE_INIT);
    }

    private void transitionTo(final State state) {
        mState = state;

        switch (mState) {
        case STATE_INIT:
        {
            resetInternalState();
        }
        break;
        case STATE_STARTUP_SYNC_COMPLETE:
        {
            // send this to prevent startup sync requests from being sent
            // from the module, so we can correctly determine if the module
            // has reset when we complete the upgrade.
            final RHAStartupSyncComplete ssc = new RHAStartupSyncComplete();
            mSerialDriver.sendMessage(ssc);

            mState = State.STATE_STARTUP_SYNC_RESPONSE;
        }
        break;
        case STATE_MODULE_INFO_REQUEST:
        {
            final RHAModuleInfoRequest frame = new RHAModuleInfoRequest();
            mSerialDriver.sendMessage(frame);

            mState = State.STATE_MODULE_INFO_RESPONSE;
        }
        break;
        case STATE_START_TRANSFER:
        {
            // first, disable serial ack for performance reasons
            final RHASerialACKConfigWrite ackConfig = new RHASerialACKConfigWrite();
            ackConfig.setSerialConfig(new SerialAckConfigEnum((byte)0x00));
            mSerialDriver.sendMessage(ackConfig);

            final RHAQueryNextImageResponse frame = new RHAQueryNextImageResponse();
            frame.setNodeId(DEFAULT_NODE_ID);
            frame.setEui64(mEui64);
            frame.setEndpoint(ALL_ENDPOINTS);
            frame.setStatus(new ZCLStatusEnum(ZCLStatusEnum.ConcreteZCLStatusEnum.SUCCESS.getCode()));
            frame.setManufacturerCode(MMB_MFG_CODE);
            frame.setImageType(DEFAULT_IMAGE_TYPE); // unused
            frame.setFileVersion(new UInt32(0)); // optional, unused
            frame.setImageSize(new UInt32(mFileSize));

            mSerialDriver.sendMessage(frame);

            mState = State.STATE_IN_TRANSFER;
        }
        break;
        case STATE_COMPLETE:
        {
            try {
                mFirmwareFile.close();
            } catch (IOException e) {} // ignore

            // re-enable serial ack
            RHASerialACKConfigWrite frame = new RHASerialACKConfigWrite();
            frame.setSerialConfig(new SerialAckConfigEnum((byte)0x01));
            mAckSequenceNumber = mSerialDriver.sendMessage(frame);
        }
        break;
        default:
            break;
        }
    }
    
    private boolean handleErrorFrame(final ARHAFrame c) {
        // only handle local bootload errors and ignore out-of-order transmission errors
        if (c instanceof RHAError) {
            final RHAError error = (RHAError)c;
            if (error.getErrorRecord().getErrorCondition().getValue() == ERROR_LOCAL_BOOTLOAD) {
                final byte subError = error.getErrorRecord().getSubError().getValue();
                if (subError != ISerialUpgradeNotification.SUBERROR_OUT_OF_ORDER_TRANSMISSION) {
                    handleUpgradeFailure(subError, true);
                    return true;
                }
            }
        }

        return false;
    }
    
    private void handleUpgradeFailure(int error, boolean resetState) {
        if (mCallback != null) {
            mCallback.notifyUpgradeFailed(error);
        }

        if (resetState) {
            resetInternalState();
        }
    }

    private void handleUpgradeSuccess() {
        if (mCallback != null) {
            mCallback.notifyUpgradeSuccess();
        }
    }
    
    private void abortUpgrade() {
        // setup the abort frame
        final ZCLStatusEnum ABORT = new ZCLStatusEnum(ZCLStatusEnum.ConcreteZCLStatusEnum.ABORT.getCode());        
        final UInt32 DEFAULT_FILE_VERSION = new UInt32(0x00000000);
        final UInt32 DEFAULT_FILE_OFFSET = new UInt32(0x00000000);

        final RHAImageBlockResponse abort = new RHAImageBlockResponse();
        abort.setNodeId(DEFAULT_NODE_ID);
        abort.setEui64(mEui64);
        abort.setEndpoint(ALL_ENDPOINTS);
        abort.setStatus(ABORT);
        abort.setManufacturerCode(MMB_MFG_CODE);
        abort.setImageType(DEFAULT_IMAGE_TYPE);
        abort.setFileVersion(DEFAULT_FILE_VERSION);
        abort.setFileOffset(DEFAULT_FILE_OFFSET);
        abort.setData(new OctetString());

        mSerialDriver.sendMessage(abort);

        resetInternalState();
    }

    private void resetInternalState() {
        mState = State.STATE_INIT;
        if (mFirmwareFile != null) {
            try {
                mFirmwareFile.close();
            } catch (IOException e) {} // ignore   
        }
        mFileSize = 0;
        mEui64 = null;
        mAckSequenceNumber = 0xff;
    }

    static final private byte UPGRADE_END_REQUEST_SUCCESS = (byte)0x00;
    static final private byte ERROR_LOCAL_BOOTLOAD = (byte)0xB0;
    static final private NodeId DEFAULT_NODE_ID = new NodeId(new byte[]{(byte)0xFF, (byte)0xFF});
    static final private UInt8 ALL_ENDPOINTS = new UInt8((short)0xFF);
    static final private UInt16 MMB_MFG_CODE = new UInt16(0x109A);
    static final private UInt16 DEFAULT_IMAGE_TYPE = new UInt16(0x0000);

    private State mState;
    private ISerialUpgradeNotification mCallback;
    private RandomAccessFile mFirmwareFile;
    private long mFileSize;
    private IEEEAddress mEui64;
    private int mAckSequenceNumber;
    final private SerialDriver mSerialDriver;
}
